import collections

from .base import ValidatorBase
from .. import Invalid
from ..schema.util import get_literal_name, get_type_name, get_primitive_name, const

# Try to load Enum type (if supported)
try:
    from enum import EnumMeta as _EnumMeta
except ImportError:
    _EnumMeta = None


class In(ValidatorBase):
    """ Validate that a value is in a collection.

    This is a plain simple `value in container` check, where `container` is a collection of literals.

    In contrast to [`Any`](#any), it does not compile its arguments into schemas,
    and hence achieves better performance.

    ```python
    from good import Schema, In

    schema = Schema(In({1, 2, 3}))

    schema(1)  #-> 1
    schema(99)
    #-> Invalid: Unsupported value: expected In(1,2,3), got 99
    ```

    The same example will work with [`Any`](#any), but slower :-)

    :param container: Collection of allowed values.

        In addition to naive tuple/list/set/dict, this can be any object that supports `in` operation.
    :type container: collections.Container
    """

    def __init__(self, container):
        assert isinstance(container, collections.Container), '`container` must support `in` operation'
        self.container = container

        # Name
        if isinstance(self.container, Map):
            self.name = self.container.name  # Inherit name
        else:
            # Format
            if isinstance(self.container, collections.Iterable):
                cs = _(u',').join(map(get_literal_name, self.container))
            else:
                cs = get_primitive_name(self.container)
            self.name = _(u'In({container})').format(container=cs)

    def __call__(self, v):
        # Test
        if v not in self.container:
            raise Invalid(_(u'Unsupported value'))

        # Okay
        return v


class Length(ValidatorBase):
    """ Validate that the provided collection has length in a certain range.

    ```python
    from good import Schema, Length

    schema = Schema(All(
        # Ensure it's a list (and not any other iterable type)
        list,
        # Validate length
        Length(max=3),
    ))
    ```

    Since mappings also have length, they can be validated as well:

    ```python
    schema = Schema({
        # Strings mapped to integers
        str: int,
        # Size = 1..3
        # Empty dicts are not allowed since `str` is implicitly `Required(str)`
        Entire: Length(max=3)
    })

    schema([1])  #-> ok
    schema([1,2,3,4])
    #-> Invalid: Too many values (3 is the most): expected Length(..3), got 4
    ```

    :param min: Minimal allowed length, or `None` to impose no limits.
    :type min: int|None
    :param max: Maximal allowed length, or `None` to impose no limits.
    :type max: int|None
    """

    def __init__(self, min=None, max=None):
        # `min` validator
        self.min_error = lambda length: Invalid(_(u'Too few values ({min} is the least)').format(min=min),
                                                get_literal_name(min), get_literal_name(length))
        self.min = min

        # `max` validator
        self.max_error = lambda length: Invalid(_(u'Too many values ({max} is the most)').format(max=max),
                                                get_literal_name(max), get_literal_name(length))
        self.max = max

        # Name
        self.name = _(u'Length({min}..{max})').format(
            min=_(u'') if min is None else min,
            max=_(u'') if max is None else max
        )

    def __call__(self, v):
        if not isinstance(v, collections.Sized):
            raise Invalid(_(u'Input is not a collection'), u'Collection', get_type_name(type(v)))

        length = len(v)

        # Validate
        if self.min is not None and length < self.min:
            raise self.min_error(length)
        if self.max is not None and length > self.max:
            raise self.max_error(length)

        # Ok
        return v


class Default(ValidatorBase):
    """ Initialize a value to a default if it's not provided.

    "Not provided" means `None`, so basically it replaces `None`s with the default:

    ```python
    from good import Schema, Any, Default

    schema = Schema(Any(
        # Accept ints
        int,
        # Replace `None` with 0
        Default(0)
    ))

    schema(1)  #-> 1
    schema(None)  #-> 0
    ```

    It raises [`Invalid`](#invalid) on all values except for `None` and `default`:

    ```python
    schema = Schema(Default(42))

    schema(42)  #-> 42
    schema(None)  #-> 42
    schema(1)
    #-> Invalid: Invalid value
    ```

    In addition, `Default` has special behavior with `Required` marker which is built into it:
    if a required key was not provided -- it's created with the default value:

    ```python
    from good import Schema, Default

    schema = Schema({
        # remember that keys are implicitly required
        'name': str,
        'age': Any(int, Default(0))
    })

    schema({'name': 'Alex'})  #-> {'name': 'Alex', 'age': 0}
    ```

    :param default: The default value to use
    """
    def __init__(self, default):
        self.default = default
        self.name = _(u'Default={default}').format(default=default)

    def __call__(self, v):
        if v is None or v is const.UNDEFINED or v == self.default:
            return self.default
        raise Invalid(_(u'Invalid value'))


class Fallback(Default):
    """ Always returns the default value.

    Works like [`Default`](#default), but does not fail on any values.

    Typical usage is to terminate [`Any`](#any) chain in case nothing worked:

    ```python
    from good import Schema, Any, Fallback

    schema = Schema(Any(
        int,
        # All non-integer numbers are replaced with `None`
        Fallback(None)
    ))
    ```

    Like [`Default`](#default), it also works with mappings.

    Internally, `Default` and `Fallback` work by feeding the schema with a special [`Undefined`](good/schema/util.py) value:
    if the schema manages to return some value without errors -- then it has the named "default behavior",
    and this validator just leverages the feature.

    A "fallback value" may be provided manually, and will work absolutely the same
    (since value schema manages to succeed even though `Undefined` was given):

    ```python
    schema = Schema({
        'name': str,
        'age': Any(int, lambda v: 42)
    })
    ```

    :param default: The value that's always returned
    """

    def __call__(self, v):
        return self.default


class Map(ValidatorBase):
    """ Convert Enumerations that map names to values.

    Supports three kinds of enumerations:

    1. Mapping.

        Provided a mapping from names to values,
        converts the input to values by mapping key:

        ```python
        from good import Schema, Map
        schema = Schema(Map({
            'RED': 0xFF0000,
            'GREEN': 0x00FF00,
            'BLUE': 0x0000FF
        }))

        schema('RED')  #-> 0xFF0000
        schema('BLACK')
        #-> Invalid: Unsupported value: expected Constant, provided BLACK
        ```

    2. Class.

        Provided a class with attributes (names) initialized with values,
        converts the input to values matching by attribute name:

        ```python
        class Colors:
            RED = 0xFF0000
            GREEN = 0x00FF00
            BLUE = 0x0000FF

        schema = Schema(Map(Colors))

        schema('RED')  #-> 0xFF0000
        schema('BLACK')
        #-> Invalid: Unsupported value: expected Colors, provided BLACK
        ```

        Note that all attributes of the class are used, except for protected (`_name`) and callables.

    3. Enum.

        Supports [Python 3.4 Enums](https://docs.python.org/3/library/enum.html)
        and the backported [enum34](https://pypi.python.org/pypi/enum34).

        Provided an enumeration, converts the input to values by name.
        In addition, enumeration value can pass through safely:

        ```python
        from enum import Enum

        class Colors(Enum):
            RED = 0xFF0000
            GREEN = 0x00FF00
            BLUE = 0x0000FF

        schema = Schema(Map(Colors))
        schema('RED')  #-> <Colors.RED: 0xFF0000>
        schema('BLACK')
        #-> Invalid: Unsupported value: expected Colors, provided BLACK
        ```

        Note that in `mode=Map.VAL` it works precisely like `Schema(Enum)`.

    In addition to the "straignt" mode (lookup by key), it supports reverse matching:

    * When `mode=Map.KEY`, does only forward matching (by key) -- the default
    * When `mode=Map.VAL`, does only reverse matching (by value)
    * When `mode=Map.BOTH`, does bidirectional matching (by key first, then by value)

    Another neat feature is that `Map` supports `in` containment checks,
    which works great together with [`In`](#in): `In(Map(enum-value))` will test if a value is convertible, but won't
    actually do the convertion.

    ```python
    from good import Schema, Map, In

    schema = Schema(In(Map(Colors)))

    schema('RED') #-> 'RED'
    schema('BLACK')
    #-> Invalid: Unsupported value, expected Colors, got BLACK
    ```

    :param enum: Enumeration: dict, object, of Enum
    :type enum: dict|type|enum.EnumMeta
    :param mode: Matching mode: one of Map.KEY, Map.VAL, Map.BOTH
    """

    KEY = 1
    VAL = 2
    BOTH = KEY|VAL

    def __init__(self, enum, mode=KEY):
        assert mode in (self.KEY, self.VAL, self.BOTH), 'Mode must be: KEY|VAL|BOTH'

        self.enum = None
        self.mode = mode
        self.name = _(u'Constant')

        self.mapping = None
        self.mapping_rev = None

        self.lookup = None
        self.rlookup = None

        # Enum (if supported)
        if _EnumMeta is not None and isinstance(enum, _EnumMeta):
            # Enum
            self.enum = enum
            self.name = enum.__name__

            # Lookups
            if self.mode & self.KEY:
                self.lookup = lambda k: k if isinstance(k, enum) else self.enum[k]
            if self.mode & self.VAL:
                self.rlookup = lambda v: self.enum(v)
        else:
            # Object?
            if not isinstance(enum, collections.Mapping):
                # Convert scalar public attributes to mapping
                self.name = enum.__name__
                enum = {k: v
                        for k, v in vars(enum).items()
                        if not (k.startswith('_') or callable(v))}

            # Mapping
            self.mapping = enum

            # Lookups
            if self.mode & self.KEY:
                self.lookup = lambda k: self.mapping[k]
            if self.mode & self.VAL:
                self.mapping_rev = {v: k for k, v in self.mapping.items()}
                self.rlookup = lambda v: self.mapping_rev[v]

    def __getitem__(self, v):
        # Try both forward and reverse lookups
        for lookup in (self.lookup, self.rlookup):
            # If enabled
            if lookup:
                try:
                    # Try to get the mapped value
                    return lookup(v)
                except Exception as e:
                    # Ok, try again
                    pass

        # Nothing worked
        raise KeyError(v)

    def __contains__(self, v):
        try:
            self[v]
        except KeyError:
            return False
        else:
            return True

    def __call__(self, v):
        try:
            return self[v]
        except KeyError:
            raise Invalid(_(u'Unsupported value'))


__all__ = ('In', 'Length', 'Default', 'Fallback', 'Map')



from __future__ import division
import six
from datetime import date, time, datetime, tzinfo, timedelta


from .base import ValidatorBase
from .. import Invalid
from ..schema.util import get_type_name


class FixedOffset(tzinfo):
    """ Fixed timezone initialized with a constant offset.

    This is only used to simulate '%z' in Python2.

    :param offset: Fixed offset, given as `timedelta` or in the "+HHMM" formatted string
    :type offset: timedelta|str
    :param name: Optional timezone name. If not provided -- is formatted as "+HHMM"
    :type name: str|None
    """

    ZERO = timedelta(0)

    @classmethod
    def parse_z(cls, offset):
        """ Parse %z offset into `timedelta` """
        assert len(offset) == 5, 'Invalid offset string format, must be "+HHMM"'
        return timedelta(hours=int(offset[:3]), minutes=int(offset[0] + offset[3:]))

    @classmethod
    def format_z(cls, offset):
        """ Format `timedelta` into %z """
        sec = offset.total_seconds()
        return '{s}{h:02d}{m:02d}'.format(s='-' if sec<0 else '+', h=abs(int(sec/3600)), m=int((sec%3600)/60))

    def __init__(self, offset, name=None):
        super(FixedOffset, self).__init__()

        # Parse
        if isinstance(offset, six.string_types):
            offset = self.parse_z(offset)

        # Create
        self._offset = offset
        self._name = name or self.format_z(self._offset)

    def utcoffset(self, dt):
        return self._offset

    def tzname(self, dt):
        return self._name

    def dst(self, dt):
        return self.ZERO

    def __repr__(self):
        return self._name


class DateTime(ValidatorBase):
    """ Validate that the input is a Python `datetime`.

    Supports the following input values:

    1. `datetime`: passthrough
    2. string: parses the string with any of the specified formats
        (see [strptime()](https://docs.python.org/3.4/library/datetime.html#strftime-and-strptime-behavior))

    ```python
    from datetime import datetime
    from good import Schema, DateTime

    schema = Schema(DateTime('%Y-%m-%d %H:%M:%S'))

    schema('2014-09-06 21:22:23')  #-> datetime.datetime(2014, 9, 6, 21, 22, 23)
    schema(datetime.now())  #-> datetime.datetime(2014, 9, 6, 21, 22, 23)
    schema('2014')
    #-> Invalid: Invalid datetime format, expected DateTime, got 2014.
    ```

    Notes on timezones:

    * If the format does not support timezones, it always returns *naive* `datetime` objects (without `tzinfo`).
    * If timezones are supported by the format (with `%z`/`%Z`),
       it returns an *aware* `datetime` objects (with `tzinfo`).
    * Since Python2 does not always support `%z` -- `DateTime` does this manually.
      Due to the limited nature of this workaround, the support for `%z` only works if it's at the end of the string!

    As a result, '00:00:00' is parsed into a *naive* datetime, and '00:00:00 +0200' results in an *aware* datetime.

    If your application wants different rules, use `localize` and `astz`:

    * `localize` argument is the default timezone to set on *naive* datetimes,
        or a callable which is applied to the input and should return adjusted `datetime`.
    * `astz` argument is the timezone to adjust the *aware* datetime to, or a callable.

    Then the generic recipe is:

    * Set `localize` to the timezone (or a callable) that you expect the user to input the datetime in
    * Set `astz` to the timezone you wish to have in the result.

    This works best with the excellent [pytz](http://pytz.sourceforge.net/) library:

    ```python
    import pytz
    from good import Schema, DateTime

    # Formats: with and without timezone
    formats = [
        '%Y-%m-%d %H:%M:%S',
        '%Y-%m-%d %H:%M:%S%z'
    ]

    # The used timezones
    UTC = pytz.timezone('UTC')
    Oslo = pytz.timezone('Europe/Oslo')

    ### Example: Use Europe/Oslo by default
    schema = Schema(DateTime(
        formats,
        localize=Oslo
    ))

    schema('2014-01-01 00:00:00')
    #-> datetime.datetime(2014, 1, 1, 0, 0, tzinfo='Europe/Oslo')
    schema('2014-01-01 00:00:00-0100')
    #-> datetime.datetime(2014, 1, 1, 0, 0, tzinfo=-0100)

    ### Example: Use Europe/Oslo by default and convert to an aware UTC
    schema = Schema(DateTime(
        formats,
        localize=Oslo,
        astz=UTC
    ))

    schema('2014-01-01 00:00:00')
    #-> datetime.datetime(2013, 12, 31, 23, 17, tzinfo=<UTC>)
    schema('2014-01-01 00:00:00-0100')
    #-> datetime.datetime(2014, 1, 1, 1, 0, tzinfo=<UTC>)

    ### Example: Use Europe/Oslo by default, convert to a naive UTC
    # This is the recommended way
    schema = Schema(DateTime(
        formats,
        localize=Oslo,
        astz=lambda v: v.astimezone(UTC).replace(tzinfo=None)
    ))

    schema('2014-01-01 00:00:00')
    #-> datetime.datetime(2013, 12, 31, 23, 17)
    schema('2014-01-01 00:00:00-0100')
    #-> datetime.datetime(2014, 1, 1, 1, 0)
    ```

    Note: to save some pain, make sure to *always* work with naive `datetimes` adjusted to UTC!
    Armin Ronacher [explains it here](http://lucumr.pocoo.org/2011/7/15/eppur-si-muove/).

    Summarizing all the above, the validation procedure is a 3-step process:

    1. Parse (only with strings)
    2. If is *naive* -- apply `localize` and make it *aware* (if `localize` is specified)
    3. If is *aware* -- apply `astz` to convert it (if `astz` is specified)

    :param formats: Supported format string, or an iterable of formats to try them all.
    :type formats: str|Iterable[str]
    :param localize: Adjust *naive* `datetimes` to a timezone, making it *aware*.

        A `tzinfo` timezone object,
        or a callable which is applied to a *naive* datetime and should return an adjusted value.

        Only called for *naive* `datetime`s.

    :type localize: datetime.tzinfo|Callable
    :param astz: Adjust *aware* `datetimes` to another timezone.

        A `tzinfo` timezone object,
        or a callable which is applied to an *aware* datetime and should return an adjusted value.

        Only called for *aware* `datetime`s, including those created by `localize`

    :type astz: datetime.tzinfo|Callable
    """

    name = get_type_name(datetime)

    # Test whether Python supports %z
    try:
        datetime.strptime('+0000', '%z')
        python_supports_z = True
    except:
        python_supports_z = False

    def __init__(self, formats, localize=None, astz=None):
        # Ensure a tuple
        self.formats = tuple([formats]
                             if isinstance(formats, six.string_types) else
                             formats)

        # Converters
        if isinstance(localize, tzinfo):
            self.localize = lambda dt: dt.replace(tzinfo=localize)
        else:
            self.localize = localize
        if isinstance(astz, tzinfo):
            self.astz = lambda dt: dt.astimezone(astz)
        else:
            self.astz = astz

        # Test converters
        if self.localize:
            assert isinstance(self.localize(datetime.now()), datetime), 'localize does not return `datetime`'
        if self.astz:
            assert isinstance(self.astz(datetime.now().replace(tzinfo=FixedOffset('+0000'))), datetime), 'astz does not return `datetime`'

    def preprocess(self, dt):
        """ Preprocess the `dt` with `localize()` and `astz()` """
        # Process
        try:  # this block should not raise errors, and if it does -- they should not be wrapped with `Invalid`
            # localize
            if self.localize and dt.tzinfo is None:
                dt = self.localize(dt)

            # astimezone
            if self.astz and dt.tzinfo is not None:
                dt = self.astz(dt)

            # Finish
            return dt
        except Exception as e:
            if isinstance(e, Invalid):
                raise
            six.reraise(RuntimeError, e)

    @classmethod
    def strptime(cls, value, format):
        """ Parse a datetime string using the provided format.

        This also emulates `%z` support on Python 2.

        :param value: Datetime string
        :type value: str
        :param format: Format to use for parsing
        :type format: str
        :rtype: datetime
        :raises ValueError: Invalid format
        :raises TypeError: Invalid input type
        """
        # Simplest case: direct parsing
        if cls.python_supports_z or '%z' not in format:
            return datetime.strptime(value, format)
        else:
            # %z emulation case
            assert format[-2:] == '%z', 'For performance, %z is only supported at the end of the string'

            # Parse
            dt = datetime.strptime(value[:-5], format[:-2])  # cutoff '%z' and '+0000'
            tz = FixedOffset(value[-5:])  # parse %z into tzinfo

            # Localize
            return dt.replace(tzinfo=tz)

    def __call__(self, v):
        # Input types
        if isinstance(v, datetime):
            dt = v
        elif not isinstance(v, (six.string_types, datetime)):
            raise Invalid(_(u'Invalid value type'), provided=get_type_name(type(v)))
        else:
            # Try all formats
            for format in self.formats:
                # Parse
                try:
                    dt = self.strptime(v, format)
                except ValueError:
                    continue
                else:
                    break
            else:
                # Nothing worked
                raise Invalid(_(u'Invalid {name} format').format(name=self.name))

        # Finish
        return self.preprocess(dt)


class Date(DateTime):
    """ Validate that the input is a Python `date`.

    Supports the following input values:

    1. `date`: passthrough
    2. `datetime`: takes the `.date()` part
    2. string: parses (see [`DateTime`](#datetime))

    ```python
    from datetime import date
    from good import Schema, Date

    schema = Schema(Date('%Y-%m-%d'))

    schema('2014-09-06')  #-> datetime.date(2014, 9, 6)
    schema(date(2014, 9, 6))  #-> datetime.date(2014, 9, 6)
    schema('2014')
    #-> Invalid: Invalid date format, expected Date, got 2014.
    ```
    """

    name = get_type_name(date)

    def __call__(self, v):
        if isinstance(v, datetime):
            return v.date()
        if isinstance(v, date):
            return v
        return super(Date, self).__call__(v).date()


class Time(DateTime):
    """ Validate that the input is a Python `time`.

    Supports the following input values:

    1. `time`: passthrough
    2. `datetime`: takes the `.timetz()` part
    2. string: parses (see [`DateTime`](#datetime))

    Since `time` is subject to timezone problems,
    make sure you've read the notes in the relevant section of [`DateTime`](#datetime) docs.
    """

    name = get_type_name(time)

    def __call__(self, v):
        if isinstance(v, time):
            v = datetime.combine(datetime.today(), v)
        return super(Time, self).__call__(v).timetz()





__all__ = ('DateTime', 'Date', 'Time')
